#
# euklid.rb
#
#  This is a ruby(1.9.1 or later) library for reading and writing 
# sav-files(HP-UX) or edx-files of Euklid(http://www.euklid-cadcam.com).
#
#
# LICENCE
#  euklid.rb is copyrighted free software by Toshimichi Oda(oda@alato.ne.jp).
#  You can redistribute it under GNU LGPL v3.
#
#
# Document
#
# module Euklid
#   module for input or output of Euklid-data
#
# Euklid.in(filename)
#   Return an array to read from sav-file 'filename'.
#
# Euklid.out(filename,data)
#   Output data to sav-file 'filename' and return filename.
#
# Euklid.edxin(filename)
#   Return an array to read from edx-file 'filename'
#
# Euklid.edxout(filename,data)
#   Output data to edx-file 'filename' and return filename.
#
# Bugs
#   1. Perhaps cannnot read sav-files of Windows.
#   2. Cannot read a sav-file including subroutine-datas.
#   3. Euklid has a bug that it cannnot write a null-list to edx-file.
#      p3d-libirary is also.
#
# Example
#   ruby -r euklid.rb -e 'Euklid.edxout("test.edx",Euklid.in("test.sav"))'
#       # convert "test.sav" into "test.edx"
#
# Changelog:
#   2010.07.29 first release
#        08.23 Euklid.edxout( ,data); data must be array.
#              Euklid.out returns filename
#              Euklid.savout2,savin2 parface with innerloops.

class Integer
  def factorial(n=self)       # factorial or falling factorial
    downto(self-n+1).inject(1,:*)
  end
  def combination(choice)     #  self! / (self-choice)! / choice!
    choice=self-choice if choice*2<self
    downto(choice+1).inject(1,:*)/(self-choice).downto(1).inject(1,:*)
  end
end
class Array
  def factorial                # product of factorials
    inject(1) {|r,i| r*(2..i).inject(1,:*)}
  end
end

module Euklid
  def Euklid.in(filename)
    open(filename,"r") {|f|
      raise "Not sav-file" unless f.read(0x200)=~/^EUKLID -- EUKLID/
      nn=f.read(4).unpack("N")[0]; nn=(nn&0x01ff)|((nn&0xfffffc00)>>1 ) ## size of data(byte)
      savin2(f,-2) }
  end

  def Euklid.out(filename,data)
    d=data.map {|d| savout2(d,true)}.join
    n=d.size+8
    open(filename,"w") {|f|
      f.write(savheader)
      f.write([(n & 0x000001ff) | (n & 0xfffffe00) <<1].pack("N"))
      f.write(d)
      f.write(["fe"].pack("H8"))  ## end of data
      f.write("\x00"*(511-(f.pos-1)%512)) }
    filename
  end

  def Euklid.edxin(filename)
    data=open(filename,"r").read
    data.gsub!(/\n/,''); data.gsub!(/""/,'')
    edxin2(data+"}")
  end

  def Euklid.edxout(filename,data)
    raise "data must be array" unless data.kind_of?(Array)
    data=edxout2(data)[1..-2]  ## remove { }
    open(filename,"w") {|f| until data.empty?
      data.match(/^.{0,78}([,}]|$)/)
      if $~.nil? || $&.count('"')%2==1
        f.write(data[0..78]+"\"\n")
        data[0..78]='"'
      else
        f.write($&+"\n")
        data.slice!(0...$~.end(0))
      end
    end}
    filename
  end

private

  def Euklid.edxin2(data,startindex=1)
    r=[]
    until data.empty? || (k=data.sub!(/^[})]/,""))
      r.push(*Array.new([0,$1.to_i-r.size-startindex].max,nil)) if  data.sub!(/^ *([-+]?[0-9]+):/,'')
      c=case
      when data.sub!(/^'(TRUE|FALSE|NIL)/      ,''); $1.downcase.to_sym
      when data.sub!(/^([-+]?[0-9][-+0-9.Ee]*)/,''); $1.downcase.to_f
      when data.sub!(/^UNKNOWN[ 0-9]*/         ,''); nil
      when data.sub!(/^"((""|[^"])*)"/         ,''); $1.sub('""','"')
      when data.sub!(/^\{0:UT1022,/            ,''); [:transformation,*delaryword(edxin2(data))]
      when data.sub!(/^\{0:UT1024,/            ,''); [:cseg,          *delaryword(edxin2(data))]
      when data.sub!(/^\{0:UT1034,/            ,''); [:parface,       *edxin2(data).map {|d| d.class==Array ? d[1,2] : d}]
      when data.sub!(/^\{0:UT1042,/            ,''); d=edxin2(data); 
        [:surf,delaryword(d[0]), *d[1..-1]]
      when data.sub!(/^\{1:/                   ,''); edxin2(data,1)
      when data.sub!(/^\{([-+]?[0-9]+:)/            ,''); edxin2(data,c=$1.to_i).unshift(:list_startindex,c)
      when data.sub!(/^\{/                     ,''); edxin2(data,1)
      when data.sub!(/^(\'\()/                 ,''); [:vector,*edxin3num(data)]
      when data.sub!(/^(\.\()/                 ,''); [:point, *edxin3num(data)]
      when data.sub!(/^(\.[Pp]\()/             ,''); [:plane ,*delaryword(edxin2(data)).flatten]
      when data.sub!(/^(\.[Ll]\()/             ,''); [:line  ,*delaryword(edxin2(data)).flatten]
      when data.sub!(/^(\.[Cc]\()/             ,''); [:circle,*delaryword(edxin2(data)).flatten]
      when data.match(/^ *[,})]/)                  ;nil
      else data.gsub!(/^[^,})]*/,""); warn ("Wrong EDX data '#{$&}'"); nil
      end
      r.push(c) if c
      data.slice!(/^ *\,/)
    end
    r
  end

  def Euklid.edxin3num(data)
     data.slice!(/^,?['.][(]/)   ## '( or .(  or ,'( or ,.(
     [data.slice!(/^([-+]?[0-9][-+0-9.Ee]*)?[,]?/).to_f,
      data.slice!(/^([-+]?[0-9][-+0-9.Ee]*)?[,]?/).to_f,
      data.slice!(/^([-+]?[0-9][-+0-9.Ee]*)?[)]/ ).to_f]
  end

  def Euklid.delaryword(ary)  ## delete type-word from 'ary'
    ary.kind_of?(Array) ?
    case ary[0]
    when :line,:plane,:circle,:point,:vector,:parface,:cseg,:transformation,:surf; ary[1..-1]
    when :list_startindex; ary[2..-1]
    else ary
    end.map {|a| delaryword(a)} : ary
  end

  def Euklid.edxout2(d)
    case d
    when :nil;   "'NIL"
    when :false; "'FALSE"
    when :true;  "'TRUE"
    when String; '"'+d+'"'
    when Numeric; "%.14g"%d
    when Array; case d[0]
      when :line;   ".L(#{edxpnt(d[1,3])},#{edxvec(d[4,3])})"
      when :plane;  ".P(#{edxpnt(d[1,3])},#{edxvec(d[4,3])})"
      when :circle; ".C(#{edxpnt(d[1,3])},#{edxvec(d[4,3])})"
      when :point;  edxpnt(d[1,3])
      when :vector; edxvec(d[1,3])
      when :parface; ["{0:UT1034",*d[1..-1].map {|i| edxpnt(i+[0])}].join(",")+"}"
      when :cseg; ["{0:UT1024,{0:"+
         d[1].zip(d[2].map {|i| i==19 ? [nil,nil,:vector] : Array.new(i)}.flatten).map {|c,v| v ? edxvec(c) : edxpnt(c)}.join(",")+"}"+
         ",{"+d[2].map {|i| "%.14g" % i}.join(",")+"}"].join(",")+"}"
      when :transformation; ["{0:UT1022",d[1..4].map {|i| edxvec(i)}].join(",")+"}"
      when :surf; "{0:UT1042,{0:"+
        d[1].map {|d1| "{0:"+d1.map {|c| edxpnt(c)}.join(",")+"}"}.join(",")+"}"+
        edxary(d[2..-1],2).sub(/^\{2:/,",")
      when :list_startindex; edxary(d[2..-1],d[1])
      else edxary(d)
     end
    end
  end

  def Euklid.savout2(d,toplevel=false)
    case d
    when :nil;  ["00"].pack("H16")
    when :false;["02"].pack("H16")
    when :true; ["03"].pack("H16")
    when Numeric; 
      r=[d].pack("G"); 
      if toplevel; r=["04"].pack("H16")+r
      else;        r[-1]=r[0]; r[0]="\x04";
      end; r
    when String; r=["18",n=((d.bytesize rescue d.size)+7)/8*2+1].pack("H8N");
         r[4]=?\x84; (0...(n-1)).each {|i| r+=(d[i*4,4]|| "").reverse.rjust(4,"\x00")};r
    when Array; case d[0]
      when :vector;["20","8c00000d",*d[1..-1]].pack("H8H8G3")
      when :point; ["21","8c00000d",*d[1..-1]].pack("H8H8G3")
      when :line;  ["24","9000000d",*d[1..-1]].pack("H8H8G6")
      when :plane; ["26","9000000d",*d[1..-1]].pack("H8H8G6")
      when :circle;["28","9000000d",*d[1..-1]].pack("H8H8G6")
      when :transformation; ["2c000000","94000021"].pack("H8H8")+[*d[1..-2].transpose,d[-1]].flatten.pack("G3x8"*3+"G3")+["3ff0"].pack("H16")
      when :parface; d=d[1..-1].each_with_object([[]]) {|a,r| (a==0)? r.push([]) : r[-1].push(a)}; sd=d.inject(0) {|r,d1|r+d1.size}
        r=["44",sd*7+d.size+1,sd*3+d.size].pack("H8N2"); r[4]=?\xa0
        i=-1; d.each {|d1| r<<[d1.size*3].pack("N")
          d1.each { r<<["02",(i+=1)*4,"04"].pack("H8NH8"); r[-8]=?\x06}}
        r<<d.flatten.pack("G*")
      when :cseg;["30000001","98000009","00000000","00000006","72","00a97c3b dummy"].pack("H8"*6)+
        ["70","00a97c35 dummy","00","00","72","bc"].pack("H8"*5+"H2")+
        [d[1].size*6+3].pack("N")[1,3]+"\x00"*4+[d[1].size*6  ].pack("N")+d[1].flatten.pack("G*")+
        ["70","ba"].pack("H8H2")+
        [d[2].size/2*2+5].pack("N")[1,3]+"\x00"*4+[d[2].size+1].pack("N")+[0,d[2],[0]*(d[2].size%2==0 ? 3 : 2)].flatten.pack("N*")
      when :surf; r=["54000001","be000023","000e0400","00",d[1].flatten.size*2+1,"00"].pack("H8"*4+"NH8")
        r[16]=?\xb8
        r+=expandcontrolpoints(dd=d[1].flatten(1),d[2],d[3]).flatten.pack("G*")
        r+="\xb4"+[d[2].size*2+1,    3,d[2].size].pack("Nnn")[1..-1]; s1=0; d[2].each {|d1| r+=[d1,s1].pack("NN"); s1+=d1*3}
        r+="\xb4"+[d[3].size*2+1,s1+=3,d[3].size].pack("Nnn")[1..-1]; s2=0; d[3].each {|d1| r+=[d1,s2].pack("NN"); s2+=d1*s1}.join("")
        r+=savary(d[4..8])+["00","00"].pack("H16H16")
        r+=(0..2).map {|i| dd.max_by {|d1| d1[i]}[i]}.pack("G3")
        r+=(0..2).map {|i| dd.min_by {|d1| d1[i]}[i]}.pack("G3")
        r+=["00"].pack("H16")
      when :list_startindex; savarystart(d[1],(d.size-2)*2+3)+savary(d[2..-1])+savaryend
      else
p "===ELSE"
                   savarystart( 1  ,(d.size-1)*2+3)+savary(d)       +savaryend
      end
    end
  end

  def Euklid.savarystart(first,size)
    "\x1c"+[first].pack("N")[1..-1]+
    "\x80"+[size ].pack("N")[1..-1]; end
  def Euklid.savary(d); d.map {|d1| savout2(d1)}.join; end
  def Euklid.savaryend; ["fc"].pack("H8"); end

  def Euklid.savheader
    ["EUKLID -- EUKLID".ljust(0x20),
    c=Time.now.strftime("%d-%b-%Y %H:%M").ljust(0x20),c,
    "EUKLID V4".ljust(0x10),
    "p4.6-6".ljust(0x10),
    RUBY_PLATFORM.ljust(0x10),   ## `uname`.chop.ljust(0x10),
    "UNKNOWN".ljust(0x10),
    "HP7".ljust(0x10),
    " "*0x10,
    "USERNAME".ljust(0x10),  ## `whoami`.chop.ljust(0x10),
    " "*0x10,
    "1234"*2+"12345678",
    " "*0x20,
    "4002".rjust(0x10),
    "5".center(0x10),
    " "*0xd0].join
  end

  def Euklid.savin2(f,datasize=-1)  # datasize <0:read until file-end  -2: toplevel
    r=[];
    while datasize!=0 && (d=f.read(4))
      case d.ord
      when 0x00,0x02,0x03; r.push([:nil,nil,:false,:true][d.ord]); f.read(4) # nil,==,false,true
      when 0x04; d+=f.read(4); if datasize==-2; d=f.read(8); else d=d[-1,1]+d[1..-2]+"\x00"; end; r.push(*d.unpack("G")); # Numeric
      when 0x18; d=f.read(4); d[0]=?\x00; n=d.unpack("N")[0]; r.push((n-1).times.map {f.read(4).reverse}.join.sub(/\x00*$/,"")) #string
      when 0x1c; 
        i=("\x00"+d[1,3]).unpack("N")[0]; i=i-0x1000000 if i>0x7fffff   # start-index
        d=f.read(4);d[0]=?\x00; n=(d.unpack("N")[0]-1)/2; r.push(i==1 ? savin2(f) : [:list_startindex,i,*savin2(f)])
      when 0x20,0x21; d+=f.read(4);r.push([[:vector,:point][d.ord-0x20],*f.read(24).unpack("G*")])  #.(),'()  # vector, point
      when 0x24,0x26,0x28; d+=f.read(4); r.push([[:line,:plane,:circle][(d.ord-0x24)/2],*f.read(8*6).unpack("G*")]) # line,plane,circle
      when 0x2c; d+=f.read(4); r.push([:transformation,*3.times.map {f.read(8*4).unpack("G3x8")}.transpose,  # transformation
                                                                     f.read(8*4).unpack("G3x8")])
      when 0x30; r.push([:cseg,(0...(f.read(4*13)[-4,4].unpack("N")[0]/6)).map {f.read(24).unpack("G3")}, # cseg
            f.read((n=f.read(4*5)[-8,4].unpack("N")[0]-1)*4).unpack("N*").map {|c| c.to_f}]); f.read(n%2==0 ? 4*3 : 4*2)
      when 0x44; s=f.read(4*f.read(4*2)[-4,4].unpack("N")[0]); r1=[]
        until s.empty?
          n=s.slice!(0,4).unpack("N")[0]
          s.slice!(0,4*n)
          r1.push((n/3).times.map {f.read(4*4).unpack("G2")})
        end
        r.push([:parface,*r1.inject {|r,i| [*r,0.0,*i]}])
      when 0x54;  # surface
        c=[(0...(("\x00"+f.read(4*5)[-7,3]).unpack("N")[0]/6)).map {f.read(8*3).unpack("G3")},
          f.read((("\x00"+f.read(8)[1,3]).unpack("N")[0]/2)*8).unpack("N*").each_slice(2).map {|a,|a},
          f.read((("\x00"+f.read(8)[1,3]).unpack("N")[0]/2)*8).unpack("N*").each_slice(2).map {|a,|a}]
        r.push([:surf,controlpoints(*c).each_slice(c[1].inject(0,:+)+1).map {|d| d },

           *c[1..2].map {|c1| c1.map {|c2| c2.to_f}},  ## change integer into float
           savin2(f,1)[0].map {|i| i==0 ? :nil : i},  ## surf[4,1] is a list of parfaces or :nil
           *savin2(f,4)])
           f.read(8*9)
      when 0xfe,0xfc; datasize=0
      end
      datasize-=1 if datasize>0
     end
    r
  end

  def Euklid.controlpoints(a,uu,vv)  # a:array of points   uu,vv:array of order-numbers
    a=a.map {|i| i.dup}
    s=uu.inject(0,:+)+1
    jj=0;(vv+[1]).each {|v| ii=0; (uu+[1]).each {|u|
      (0...v).each {|j| (0...u).each {|i| (0..2).each {|x|
        a[c=s*(jj+j)+ii+i][x]=a[c][x]/u.combination(i)/v.combination(j)-
          (0..i).inject(0) {|ri,i1| ri+
          (0..j).inject(0) {|rj,j1| rj+((i1==i && j1==j) ? 0 :
          a[s*(jj+j1)+ii+i1][x]*j.combination(j1)*(1-(i+j+i1+j1)%2*2))}*i.combination(i1)}
      }}}
    ii+=u}; jj+=v}
    a
  end

  def Euklid.expandcontrolpoints(a,uu,vv) # reverse of controlpoints
    s=uu.inject(0,:+)+1; r=[]
    jj=0; (vv+[1]).each {|v| (0...v).each {|j|
    ii=0; (uu+[1]).each {|u| (0...u).each {|i| r.push((0..2).map {|x|
      (0..i).inject(0) {|ri,i1| ri+
      (0..j).inject(0) {|rj,j1| rj+a[s*(jj+j1)+ii+i1][x]/[j1,j-j1].factorial*(1-(i+j+i1+j1)%2*2)}/[i1,i-i1].factorial}*
        u.to_i.factorial(i)*v.to_i.factorial(j)})}
    ii+=u}}; jj+=v}
    r
  end

  def Euklid.edxary(d,startindex=1)
    r=[];  nilnil=false;
    d.each_with_index {|d1,i|
      c=(nilnil==true && d1!=:nil ? "%d:"%[startindex+i] : "")+
        ((nilnil=(d1==:nil)) ? "" : edxout2(d1))
      r.push(c) unless c.empty? }
   r.push("%d:" % (startindex+d.size-1)) if nilnil
   (startindex!=1||d[0]==:nil ? "{%d:" % startindex : "{")+r.join(",")+"}"
  end

  def Euklid.edxpnt(xyz); (".(%.14g,%.14g,%.14g)" % xyz).gsub(/([(,])0(?=[),])/,'\1').sub(/,+[)]/,")"); end
  def Euklid.edxvec(xyz); ("'(%.14g,%.14g,%.14g)" % xyz).gsub(/([(,])0(?=[),])/,'\1').sub(/,+[)]/,")"); end

end  ## end-of-module-Euklid
